/*
 *	File:		gr_arbview.c
 *	Contents:	View level routines for blocked data for graphics module
 *
 *  Contributed by R. Hugo Patterson at Carnegie Mellon University
 */

#include "gr_com.h"

extern	void	gr_MatInverse();
extern	void	td_VolumeInterp();
extern  void	gr_Box2Data();

#define BDS(_x_,_y_,_z_) \
    bDs [_x_>>xshift] [_y_>>yshift] [_z_>>zshift].data[((_x_&xmask)<<yzshift) + ((_y_&ymask)<<zshift) + (_z_&zmask)]


static int
IncludeBlock(bd)
A_BlkDes_t *bd;
{
	if (!bd->flag) {
		bd->flag = 1;
		if (bd->data == NULL)
			return 1;
	}
	return 0;
}


/*
 * This routine returns an array of pixels for the arbitrary slice selected. 
 * It uses a three step algorithm.  First, the pixels are mapped to data set
 * coordinates and the containing blocks are marked as needed.  A list of
 * coordinates is maintained to avoid repeating the mapping.  Second, the
 * needed blocks are read in by calling td_HdfLoadArb.  Lastly, a pixel is
 * generated for each coordinate on the list.
 */
char
*gr_ViewgetArbData(box,scale,viewType,bossWin)
A_Box_t *box;
int		scale;
int		viewType;
A_BossWind_t	*bossWin;
{
	A_BlkData_t *arDat=&(bossWin->data->blkData);
	A_BlkDes_t	***bDs=bossWin->data->blkData.blkDes;
	float32		val;
	float32		min=(float32)bossWin->data->min;
	float32		max=(float32)bossWin->data->max;
	float32		frac=(float32)bossWin->data->rangeFrac;
	Boolean		interp;
	double		xd,yd,zd,i,j,incr;
	char		*data;
	int			count, pixcount, blksneeded, x,y,z, x1, y1, z1;
	int			xdim,ydim,zdim;
	int			xstart,xend,ystart,yend;
	int			xshift, yshift,zshift, yzshift, xmask,ymask,zmask;
	unsigned char nColors = (unsigned char)gr_color.nColors;
	struct coords {
		float32 x,y,z;
	} *coord;

	if (viewType == 1)
	{
		xstart = (int)box->boundxmin;
		xend   = (int)box->boundxmax;
		ystart = (int)box->boundymin;
		yend   = (int)box->boundymax;
	}
	else
	{
		xstart = (int)box->xmin;
		xend   = (int)box->xmax;
		ystart = (int)box->ymin;
		yend   = (int)box->ymax;
	}

	xdim   = box->xdim-1;
	ydim   = box->ydim-1;
	zdim   = box->zdim-1;
	incr   = (double)1.0/(double)scale;
	box->xpicsize = (int)((xend-xstart)*scale);
	box->ypicsize = (int)((yend-ystart)*scale);
	if ((box->xpicsize == 0) || (box->ypicsize == 0))
	{
		gr_TextMsgOut("Warning:\t Please select a valid slice first!\n");
		return(NULL);
	}
	if ((data  = td_Malloc1D(box->xpicsize,box->ypicsize,
				 (long)sizeof(unsigned char),"td_Malloc1D")) == NULL)
		return(NULL);
	if ((coord  = (struct coords *)td_Malloc1D(box->xpicsize,box->ypicsize,
				 (long)sizeof(struct coords),"td_Malloc1D")) == NULL) {
	    td_Free(data);
		return(NULL);
	}
	
	gr_MatInverse(box->matrix,box->invmatrix);

	interp = bossWin->arbWin->interp;
	xshift = arDat->logblksizes[0];
	yshift = arDat->logblksizes[1];
    zshift = arDat->logblksizes[2];
	yzshift = yshift + zshift;
	xmask = (1 << xshift) - 1;
	ymask = (1 << yshift) - 1;
	zmask = (1 << zshift) - 1;

	/* go through mapping process to determine what data is needed.
	   This is a brute force approach that a little thought should
	   make go much faster. */
	pixcount = 0;
	blksneeded = 0;
	for (i=(double)yend;i>(double)ystart;i-=incr)
	{
		for (j=(double)xstart;j<(double)xend;j+=incr)
		{
			gr_Box2Data(box,j,i,box->depth,&xd,&yd,&zd);
			coord[pixcount].x = (float32)xd; /* cache coordinates */
			x = (int)coord[pixcount].x;
			coord[pixcount].y = (float32)yd;
			y = (int)coord[pixcount].y;
			coord[pixcount].z = (float32)zd;
			z = (int)coord[pixcount].z;

			if (((x >= 0) && (x < xdim)) &&
				((y >= 0) && (y < ydim)) &&
				((z >= 0) && (z < zdim)))
			{
				if (interp == TRUE) {
					x1 = (x+1) >> xshift;
					x = x >> xshift;
					y1 = (y+1) >> yshift;
					y = y >> yshift;
					z1 = (z+1) >> zshift;
					z = z >> zshift;

					blksneeded += IncludeBlock(&bDs[x][y][z]);
					if (x1 != x) {
						blksneeded += IncludeBlock(&bDs[x1][y][z]);
						if (y1 != y) {
							blksneeded += IncludeBlock(&bDs[x][y1][z]);
							blksneeded += IncludeBlock(&bDs[x1][y1][z]);
							if (z1 != z) {
								blksneeded += IncludeBlock(&bDs[x][y][z1]);
								blksneeded += IncludeBlock(&bDs[x1][y][z1]);
								blksneeded += IncludeBlock(&bDs[x][y1][z1]);
								blksneeded += IncludeBlock(&bDs[x1][y1][z1]);
							}
						} 
						else 
							if (z1 != z) {
								blksneeded += IncludeBlock(&bDs[x][y][z1]);
								blksneeded += IncludeBlock(&bDs[x1][y][z1]);
							}
					} 
					else {
						if (y1 != y) {
							blksneeded += IncludeBlock(&bDs[x][y1][z]);
							if (z1 != z) {
								blksneeded += IncludeBlock(&bDs[x][y][z1]);
								blksneeded += IncludeBlock(&bDs[x][y1][z1]);
							}
						} else
							if (z1 != z)
								blksneeded += IncludeBlock(&bDs[x][y][z1]);
					}
				}
				else {
					x = x >> xshift;
					y = y >> yshift;
					z = z >> zshift;
					blksneeded += IncludeBlock(&bDs[x][y][z]);
				}
			}
			else
				coord[pixcount].x = (float32)(xdim+1);
			pixcount++;
		}
	}	

	/* read in data */
	if (blksneeded) {
		if (td_HdfLoadArb(blksneeded, bossWin->data) == -1)
		{
			sprintf(msg,"***HDF ERROR: Cannot load SDS %d.\n",
					bossWin->data->num);
			gr_TextMsgOut(msg);
			td_Free(data);
			td_Free(coord);
			return(NULL);
		}
	}

	/* use data to generate pixels */
	for (count=0;count<pixcount;count++)
	{
		xd = coord[count].x;
		x = (int)coord[count].x;
		if (x < xdim)		/* x,y,z in bounds */
		{
			yd = coord[count].y;
			zd = coord[count].z;
			y = (int)coord[count].y;
			z = (int)coord[count].z;
			if (interp == TRUE)
			{
				xd=xd-(double)x; yd=yd-(double)y; zd=zd-(double)z;
				td_VolumeInterp(&val,BDS(x,y,z),BDS(x+1,y,z),
								BDS(x+1,y+1,z),BDS(x,y+1,z),BDS(x,y,z+1),
								BDS(x+1,y,z+1),BDS(x+1,y+1,z+1),
								BDS(x,y+1,z+1),xd,yd,zd);
			}
			else
				val = BDS(x,y,z);
			data[count]=XDS_CNV2PIX(val,min,max,frac,nColors);
		}
		else
			data[count]=0;
	}	
	td_Free(coord);
	return(data);
}
